/*++
/* NAME
/*	lmtp_connect 3
/* SUMMARY
/*	connect to LMTP server
/* SYNOPSIS
/*	#include "lmtp.h"
/*
/*	LMTP_SESSION *lmtp_connect(destination, why)
/*	char	*destination;
/*	VSTRING	*why;
/* DESCRIPTION
/*	This module implements LMTP connection management.
/*
/*	lmtp_connect() attempts to establish an LMTP session.
/*
/*	The destination is one of the following:
/* .IP unix:address
/*	Connect to a UNIX-domain service. The destination is a pathname.
/*	Beware: UNIX-domain sockets are flakey on Solaris, at last up to
/*	and including Solaris 7.0.
/* .IP inet:address
/*	Connect to an IPV4 service.
/*	The destination is either a host name or a numeric address.
/*	Symbolic or numeric service port information may be appended,
/*	separated by a colon (":").
/*
/*	Numerical address information should always be quoted with `[]'.
/* .PP
/*	When no transport is specified, "inet" is assumed.
/* DIAGNOSTICS
/*	This routine either returns an LMTP_SESSION pointer, or
/*	returns a null pointer and set the \fIlmtp_errno\fR
/*	global variable accordingly:
/* .IP LMTP_RETRY
/*	The connection attempt failed, but should be retried later.
/* .IP LMTP_FAIL
/*	The connection attempt failed.
/* .PP
/*	In addition, a textual description of the error is made available
/*	via the \fIwhy\fR argument.
/* SEE ALSO
/*	lmtp_proto(3) LMTP client protocol
/* LICENSE
/* .ad
/* .fi
/*	The Secure Mailer license must be distributed with this software.
/* AUTHOR(S)
/*	Wietse Venema
/*	IBM T.J. Watson Research
/*	P.O. Box 704
/*	Yorktown Heights, NY 10598, USA
/*
/*	Alterations for LMTP by:
/*	Philip A. Prindeville
/*	Mirapoint, Inc.
/*	USA.
/*
/*	Additional work on LMTP by:
/*	Amos Gouaux
/*	University of Texas at Dallas
/*	P.O. Box 830688, MC34
/*	Richardson, TX 75083, USA
/*--*/

/* System library. */

#include <sys_defs.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <netdb.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>

#ifdef STRCASECMP_IN_STRINGS_H
#include <strings.h>
#endif

/* Utility library. */

#include <msg.h>
#include <vstream.h>
#include <vstring.h>
#include <split_at.h>
#include <mymalloc.h>
#include <iostuff.h>
#include <timed_connect.h>
#include <stringops.h>
#include <host_port.h>
#include <sane_connect.h>
#include <inet_addr_list.h>

/* Global library. */

#include <mail_params.h>
#include <mail_proto.h>
#include <own_inet_addr.h>

/* DNS library. */

#include <dns.h>

#ifdef INET6
#define GAI_STRERROR(error) \
	((error == EAI_SYSTEM) ? strerror(errno) : gai_strerror(error))
#endif
	
/* Application-specific. */

#include "lmtp.h"
#include "lmtp_addr.h"

 /*
  * Forward declaration.
  */
static LMTP_SESSION *lmtp_connect_sock(int, struct sockaddr *, int,
				               const char *, const char *,
				               const char *, VSTRING *);

/* lmtp_connect_unix - connect to UNIX-domain address */

static LMTP_SESSION *lmtp_connect_unix(const char *addr,
			              const char *destination, VSTRING *why)
{
#undef sun
    char   *myname = "lmtp_connect_unix";
    struct sockaddr_un sun;
    int     len = strlen(addr);
    int     sock;

    /*
     * Sanity checks.
     */
    if (len >= (int) sizeof(sun.sun_path)) {
	msg_warn("unix-domain name too long: %s", addr);
	lmtp_errno = LMTP_RETRY;
	return (0);
    }

    /*
     * Initialize.
     */
    memset((char *) &sun, 0, sizeof(sun));
    sun.sun_family = AF_UNIX;
#ifdef HAS_SUN_LEN
    sun.sun_len = len + 1;
#endif
    memcpy(sun.sun_path, addr, len + 1);

    /*
     * Create a client socket.
     */
    if ((sock = socket(AF_UNIX, SOCK_STREAM, 0)) < 0)
	msg_fatal("%s: socket: %m", myname);

    /*
     * Connect to the LMTP server.
     */
    if (msg_verbose)
	msg_info("%s: trying: %s...", myname, addr);

    return (lmtp_connect_sock(sock, (struct sockaddr *) & sun, sizeof(sun),
			      addr, addr, destination, why));
}

/* lmtp_force_bind: bind() address */

static void lmtp_force_bind(const char *bind_addr,
			    const char *bind_var,
			    int sock,
			    int af)
{
    /*
     * If the bind() call fails, this is considered a non-fatal error.
     * All address conversion errors are fatal.
     */
    char   *myname = "lmtp_force_bind";
#ifdef INET6
    char    hbuf[NI_MAXHOST];
    int     aierr;
    struct addrinfo hints, *res;

    memset(&hints, 0, sizeof(hints));
    hints.ai_family = af;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = AI_PASSIVE | AI_NUMERICHOST;
    snprintf(hbuf, sizeof(hbuf), "%s", bind_addr);
    aierr = getaddrinfo(hbuf, NULL, &hints, &res);
    if (aierr == EAI_NONAME)
	msg_fatal("%s: bad %s parameter: \"%s\"",
		  myname, bind_var, bind_addr);
    if (aierr != 0) {
	if (msg_verbose)
	    msg_warn("%s: getaddrinfo(%s): %s",
		     myname, hbuf, GAI_STRERROR(aierr));
	return;
    }
    aierr = getnameinfo(res->ai_addr, res->ai_addrlen, hbuf, sizeof(hbuf),
			NULL, 0, NI_NUMERICHOST | NI_WITHSCOPEID);
    if (aierr != 0) {
	msg_warn("%s: getnameinfo(): %s",
		 myname, GAI_STRERROR(aierr));
	freeaddrinfo(res);
	return;
    }
    if (bind(sock, res->ai_addr, res->ai_addrlen) < 0)
	msg_warn("%s: bind %s: %m", myname, hbuf);
    else if (msg_verbose)
	msg_info("%s: bind %s", myname, hbuf);
    freeaddrinfo(res);
#else /* INET6 */
    struct sockaddr_in sin;

    memset(&sin, 0, sizeof(sin));
    sin.sin_family = AF_INET;
#ifdef HAS_SA_LEN
    sin.sin_len = sizeof(sin);
#endif
    sin.sin_addr.s_addr = inet_addr(bind_addr);
    if (sin.sin_addr.s_addr == INADDR_NONE) {
	msg_fatal("%s: bad %s parameter: \"%s\"",
		  myname, bind_var, bind_addr);
	return;
    }
    if (bind(sock, (struct sockaddr *)&sin, sizeof(sin)) < 0)
	msg_warn("%s: bind %s: %m", myname, inet_ntoa(sin.sin_addr));
    else if (msg_verbose)
	msg_info("%s: bind %s", myname, inet_ntoa(sin.sin_addr));
#endif /* INET6 */
}

/* lmtp_virtual_bind - bind() when acting as virtual host */

static void lmtp_virtual_bind(int sock, int af)
{
    char    *myname = "lmtp_virtual_bind";
    INET_ADDR_LIST *addr_list;
    int     count;

#ifdef INET6
    int     i;
    char    hbuf[NI_MAXHOST];
    int     aierr;
    struct sockaddr *sa;
    struct addrinfo hints, *loopback = NULL, *res = NULL;

    /*
     * Check whether we are acting as a virtual host
     */
    count = 0;
    addr_list = own_inet_addr_list();
    for (i = 0; count < 2 && i < addr_list->used; i++)
	if (((struct sockaddr *)&addr_list->addrs[i])->sa_family == af)
	    count++;
    if (count != 1)
	return;

    /*
     * Bind the source address.
     */
    memset(&hints, 0, sizeof(hints));
    hints.ai_family = af;
    hints.ai_socktype = SOCK_STREAM;
    aierr = getaddrinfo(NULL, "0", &hints, &loopback);
    if (aierr != 0) {
	loopback = NULL;
	msg_warn("%s: getaddrinfo(\"0\"): %s",
		 myname, GAI_STRERROR(aierr));
    }

    sa = (struct sockaddr *)&addr_list->addrs[i - 1];
    aierr = getnameinfo(sa, SA_LEN(sa), hbuf, sizeof(hbuf),
			NULL, 0, NI_NUMERICHOST | NI_WITHSCOPEID);
    if (aierr != 0)
	msg_fatal("%s: getnameinfo() (AF=%d): %s",
		  myname, af, GAI_STRERROR(aierr));

    memset(&hints, 0, sizeof(hints));
    hints.ai_family = af;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = AI_NUMERICHOST | AI_PASSIVE;
    aierr = getaddrinfo(hbuf, NULL, &hints, &res);
    if (aierr != 0)
	msg_fatal("%s: getaddrinfo(\"%s\"): %s",
		  myname, hbuf, GAI_STRERROR(aierr));

    if (res->ai_addrlen != loopback->ai_addrlen
	|| memcmp(res->ai_addr, loopback->ai_addr, res->ai_addrlen) != 0) {
	if (bind(sock, res->ai_addr, res->ai_addrlen) < 0)
	    msg_warn("%s: bind %s: %m", myname, hbuf);
	else if (msg_verbose)
	    msg_info("%s: bind %s", myname, hbuf);
    } else if (msg_verbose) {
	msg_info("%s: not calling bind(): unusable source "
		 "address from \"%s\"", myname, hbuf);
    }
    if (res)
	freeaddrinfo(res);
    if (loopback)
	freeaddrinfo(loopback);

#else /* INET6 */

    struct sockaddr_in sin;
    unsigned long inaddr;	/*XXX BAD!*/

    /*
     * Check whether we are acting as a virtual host
     */
    addr_list = own_inet_addr_list();
    count = addr_list->used;
    if (count != 1)
	return;

    /*
     * Bind the source address.
     */
    memset(&sin, 0, sizeof(sin));
    sin.sin_family = AF_INET;
#ifdef HAS_SA_LEN
    sin.sin_len = sizeof(sin);
#endif
    memcpy((char *) &sin.sin_addr, addr_list->addrs, sizeof(sin.sin_addr));
    inaddr = (unsigned long)ntohl(sin.sin_addr.s_addr);
    if (!IN_CLASSA(inaddr)
	|| !(((inaddr & IN_CLASSA_NET) >> IN_CLASSA_NSHIFT) == IN_LOOPBACKNET)) {
	if (bind(sock, (struct sockaddr *) & sin, sizeof(sin)) < 0)
	    msg_warn("%s: bind %s: %m", myname, inet_ntoa(sin.sin_addr));
	else if (msg_verbose)
	    msg_info("%s: bind %s", myname, inet_ntoa(sin.sin_addr));
    }
#endif /* INET6 */
}

/* lmtp_connect_addr - connect to explicit address */

static LMTP_SESSION *lmtp_connect_addr(DNS_RR *addr, unsigned port,
			              const char *destination, VSTRING *why)
{
    char   *myname = "lmtp_connect_addr";
#ifdef INET6
    struct sockaddr_storage ss;
#else
    struct sockaddr ss;
#endif
    struct sockaddr *sa;
    struct sockaddr_in *sin;
#ifdef INET6
    struct sockaddr_in6 *sin6;
#endif
    SOCKADDR_SIZE salen;
#ifdef INET6
    char hbuf[NI_MAXHOST];
#else
    char hbuf[sizeof("255.255.255.255") + 1];
#endif
    int     sock = -1;
    INET_ADDR_LIST *addr_list;
    char    *bind_addr;
    char    *bind_var;
#ifdef INET6
    char    *addr6_ptr = NULL;
#endif

    sa = (struct sockaddr *)&ss;
    sin = (struct sockaddr_in *)&ss;
#ifdef INET6
    sin6 = (struct sockaddr_in6 *)&ss;
#endif

    /*
     * Sanity checks.
     */
#ifdef INET6
    if (((addr->type==T_A) && (addr->data_len > sizeof(sin->sin_addr))) ||
	((addr->type==T_AAAA) && (addr->data_len > sizeof(sin6->sin6_addr))))
#else
    if (addr->data_len > sizeof(sin->sin_addr))
#endif
    {
	msg_warn("%s: skip address with length %d", myname, addr->data_len);
	lmtp_errno = LMTP_RETRY;
	return (0);
    }

    /*
     * Initialize.
     */
    switch (addr->type) {
#ifdef INET6
    case T_AAAA:
	bind_addr = "";
	bind_var = VAR_LMTP_BIND_ADDR6;
	if (*var_lmtp_bind_addr6) {
	    addr6_ptr = mystrdup(var_lmtp_bind_addr6);
	    if (*addr6_ptr == '[' && addr6_ptr[strlen(addr6_ptr) - 1] == ']') {
		addr6_ptr[strlen(addr6_ptr) - 1] = 0;
		bind_addr = addr6_ptr + 1;
	    } else {
		msg_warn("%s: skip incorrectly bracketed IPv6 address in %s",
		    myname, VAR_LMTP_BIND_ADDR6);
	    }
	}
	memset(sin6, 0, sizeof(*sin6));
	sin6->sin6_family = AF_INET6;
	salen = sizeof(*sin6);
	break;
#endif
    default: /* T_A: */
	bind_addr = var_lmtp_bind_addr;
	bind_var = VAR_SMTP_BIND_ADDR;
	memset(sin, 0, sizeof(*sin));
	sin->sin_family = AF_INET;
	salen = sizeof(*sin);
	break;
    };
#ifdef HAS_SALEN
    sa->sa_len = salen;
#endif

    if ((sock = socket(sa->sa_family, SOCK_STREAM, 0)) < 0)
	msg_fatal("%s: socket: %m", myname);

    /*
     * Allow the sysadmin to specify the source address
     */

    if (bind_addr && *bind_addr) {
	lmtp_force_bind(bind_addr, bind_var, sock, sa->sa_family);
#ifdef INET6
	if (addr6_ptr)
	    myfree(addr6_ptr);
#endif
    } else {
	/*
	 * When running as a virtual host, bind to the virtual interface so that
	 * the mail appears to come from the "right" machine address.
	 */
	lmtp_virtual_bind(sock, sa->sa_family);
    }

    /*
     * Connect to the LMTP server.
     */
    switch (addr->type) {
#ifdef INET6
    case T_AAAA:
	/* XXX scope-unfriendly */
	memset(sin6, 0, sizeof(*sin6));
	sin6->sin6_port = port;
	sin6->sin6_family = AF_INET6;
	salen = sizeof(*sin6);
	memcpy(&sin6->sin6_addr, addr->data, sizeof(sin6->sin6_addr));
	inet_ntop(AF_INET6, &sin6->sin6_addr, hbuf, sizeof(hbuf));
	break;
#endif
    default: /* T_A: */
	memset(sin, 0, sizeof(*sin));
	sin->sin_port = port;
	sin->sin_family = AF_INET;
	salen = sizeof(*sin);
	memcpy(&sin->sin_addr, addr->data, sizeof(sin->sin_addr));
	inet_ntop(AF_INET, &sin->sin_addr, hbuf, sizeof(hbuf));
	break;
    }
#ifdef HAS_SA_LEN
    sa->sa_len = salen;
#endif

    if (msg_verbose)
	msg_info("%s: trying: %s[%s] port %d...",
		 myname, addr->name, hbuf, ntohs(port));

    return (lmtp_connect_sock(sock, (struct sockaddr *)sa, salen,
			      addr->name, hbuf, destination, why));
}

/* lmtp_connect_sock - connect a socket over some transport */

static LMTP_SESSION *lmtp_connect_sock(int sock, struct sockaddr * sa, int len,
				         const char *name, const char *addr,
			              const char *destination, VSTRING *why)
{
    int     conn_stat;
    int     saved_errno;
    VSTREAM *stream;
    int     ch;

    if (var_lmtp_conn_tmout > 0) {
	non_blocking(sock, NON_BLOCKING);
	conn_stat = timed_connect(sock, sa, len, var_lmtp_conn_tmout);
	saved_errno = errno;
	non_blocking(sock, BLOCKING);
	errno = saved_errno;
    } else {
	conn_stat = sane_connect(sock, sa, len);
    }
    if (conn_stat < 0) {
	vstring_sprintf(why, "connect to %s[%s]: %m",
			name, addr);
	lmtp_errno = LMTP_RETRY;
	close(sock);
	return (0);
    }

    /*
     * Skip this host if it takes no action within some time limit.
     */
    if (read_wait(sock, var_lmtp_lhlo_tmout) < 0) {
	vstring_sprintf(why, "connect to %s[%s]: read timeout",
			name, addr);
	lmtp_errno = LMTP_RETRY;
	close(sock);
	return (0);
    }

    /*
     * Skip this host if it disconnects without talking to us.
     */
    stream = vstream_fdopen(sock, O_RDWR);
    if ((ch = VSTREAM_GETC(stream)) == VSTREAM_EOF) {
	vstring_sprintf(why, "connect to %s[%s]: server dropped connection without sending the initial greeting",
			name, addr);
	lmtp_errno = LMTP_RETRY;
	vstream_fclose(stream);
	return (0);
    }
    vstream_ungetc(stream, ch);

    /*
     * Skip this host if it sends a 4xx or 5xx greeting.
     */
    if (ch == '4' || ch == '5') {
	vstring_sprintf(why, "connect to %s[%s]: server refused mail service",
			name, addr);
	lmtp_errno = LMTP_RETRY;
	vstream_fclose(stream);
	return (0);
    }
    return (lmtp_session_alloc(stream, name, addr, destination));
}

/* lmtp_connect_host - direct connection to host */

static LMTP_SESSION *lmtp_connect_host(char *host, unsigned port,
			              const char *destination, VSTRING *why)
{
    LMTP_SESSION *session = 0;
    DNS_RR *addr_list;
    DNS_RR *addr;

    /*
     * Try each address in the specified order until we find one that works.
     * The addresses belong to the same A record, so we have no information
     * on what address is "best".
     */
    addr_list = lmtp_host_addr(host, why);
    for (addr = addr_list; addr; addr = addr->next) {
	if ((session = lmtp_connect_addr(addr, port, destination, why)) != 0) {
	    break;
	}
    }
    dns_rr_free(addr_list);
    return (session);
}

/* lmtp_parse_destination - parse destination */

static char *lmtp_parse_destination(const char *destination, char *def_service,
				            char **hostp, unsigned *portp)
{
    char   *myname = "lmtp_parse_destination";
    char   *buf = mystrdup(destination);
    char   *service;
    struct servent *sp;
    char   *protocol = "tcp";		/* XXX configurable? */
    unsigned port;
    const char *err;

    if (msg_verbose)
	msg_info("%s: %s %s", myname, destination, def_service);

    /*
     * Parse the host/port information. We're working with a copy of the
     * destination argument so the parsing can be destructive.
     */
    if ((err = host_port(buf, hostp, &service, def_service)) != 0)
	msg_fatal("%s in LMTP server description: %s", err, destination);

    /*
     * Convert service to port number, network byte order. Since most folks
     * aren't going to have lmtp defined as a service, use a default value
     * instead of just blowing up.
     */
    if (alldig(service) && (port = atoi(service)) != 0)
	*portp = htons(port);
    else if ((sp = getservbyname(service, protocol)) != 0)
	*portp = sp->s_port;
    else
	*portp = htons(var_lmtp_tcp_port);
    return (buf);
}

/* lmtp_connect - establish LMTP connection */

LMTP_SESSION *lmtp_connect(const char *destination, VSTRING *why)
{
    char   *myname = "lmtp_connect";
    LMTP_SESSION *session;
    char   *dest_buf;
    char   *host;
    unsigned port;
    char   *def_service = "lmtp";	/* XXX configurable? */

    /*
     * Connect to the LMTP server.
     * 
     * XXX Ad-hoc transport parsing and connection management. Some or all
     * should be moved away to a reusable library routine so that every
     * program benefits from it.
     * 
     * XXX Should transform destination into canonical form (unix:/path or
     * inet:host:port before entering it into the connection cache. See also
     * the connection cache lookup code in lmtp.c.
     */
    if (strncmp(destination, "unix:", 5) == 0)
	return (lmtp_connect_unix(destination + 5, destination, why));
    if (strncmp(destination, "inet:", 5) == 0)
	dest_buf = lmtp_parse_destination(destination + 5, def_service,
					  &host, &port);
    else
	dest_buf = lmtp_parse_destination(destination, def_service,
					  &host, &port);
    if (msg_verbose)
	msg_info("%s: connecting to %s port %d", myname, host, ntohs(port));
    session = lmtp_connect_host(host, port, destination, why);
    myfree(dest_buf);
    return (session);
}
