/*++
/* NAME
/*	smtpd_peer 3
/* SUMMARY
/*	look up peer name/address information
/* SYNOPSIS
/*	#include "smtpd.h"
/*
/*	void	smtpd_peer_init(state)
/*	SMTPD_STATE *state;
/*
/*	void	smtpd_peer_reset(state)
/*	SMTPD_STATE *state;
/* DESCRIPTION
/*	The smtpd_peer_init() routine attempts to produce a printable
/*	version of the peer name and address of the specified socket.
/*	Where information is unavailable, the name and/or address
/*	are set to "unknown".
/*
/*	smtpd_peer_init() updates the following fields:
/* .IP name
/*	The client hostname. An unknown name is represented by the
/*	string "unknown".
/* .IP addr
/*	Printable representation of the client address.
/* .IP namaddr
/*	String of the form: "name[addr]".
/* .IP peer_code
/*	The peer_code result field specifies how the client name
/*	information should be interpreted:
/* .RS
/* .IP 2
/*	Both name lookup and name verification succeeded.
/* .IP 4
/*	The name lookup or name verification failed with a recoverable
/*	error (no address->name mapping or no name->address mapping).
/* .IP 5
/*	The name lookup or verification failed with an unrecoverable
/*	error (no address->name mapping, bad hostname syntax, no
/*	name->address mapping, client address not listed for hostname).
/* .RE
/* .PP
/*	smtpd_peer_reset() releases memory allocated by smtpd_peer_init().
/* LICENSE
/* .ad
/* .fi
/*	The Secure Mailer license must be distributed with this software.
/* AUTHOR(S)
/*	Wietse Venema
/*	IBM T.J. Watson Research
/*	P.O. Box 704
/*	Yorktown Heights, NY 10598, USA
/*--*/

/* System library. */

#include <sys_defs.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>			/* strerror() */
#include <errno.h>
#include <netdb.h>
#include <string.h>

/* Utility library. */

#include <msg.h>
#include <mymalloc.h>
#include <valid_hostname.h>
#include <stringops.h>
#ifdef INET6
#include <inet_addr_list.h>	/* for NI_WITHSCOPEID */
#endif

/* Global library. */

#include <mail_proto.h>

 /*
  * Older systems don't have h_errno. Even modern systems don't have
  * hstrerror().
  */
#ifdef NO_HERRNO

static int h_errno = TRY_AGAIN;

#define  HSTRERROR(err) "Host not found"

#else

#define  HSTRERROR(err) (\
	err == TRY_AGAIN ? "Host not found, try again" : \
	err == HOST_NOT_FOUND ? "Host not found" : \
	err == NO_DATA ? "Host name has no address" : \
	err == NO_RECOVERY ? "Name server failure" : \
	strerror(errno) \
    )
#endif

#ifdef INET6
#define GAI_STRERROR(error) \
	((error == EAI_SYSTEM) ? strerror(errno) : gai_strerror(error))
#endif
	
/* Application-specific. */

#include "smtpd.h"

/* smtpd_peer_init - initialize peer information */

void    smtpd_peer_init(SMTPD_STATE *state)
{
    char  *myname = "smtpd_peer_init";
#ifdef INET6
    struct sockaddr_storage ss;
#else
    struct sockaddr ss;
    struct in_addr *in;
    struct hostent *hp;
#endif
    struct sockaddr *sa;
    SOCKADDR_SIZE len;

    sa = (struct sockaddr *)&ss;
    len = sizeof(ss);

    /*
     * Look up the peer address information.
     */
    if (getpeername(vstream_fileno(state->client), sa, &len) >= 0) {
	errno = 0;
    }

    /*
     * If peer went away, give up.
     */
    if (errno == ECONNRESET || errno == ECONNABORTED) {
	state->name = mystrdup(CLIENT_NAME_UNKNOWN);
	state->addr = mystrdup(CLIENT_ADDR_UNKNOWN);
	state->peer_code = SMTPD_PEER_CODE_PERM;
    }

    /*
     * Look up and "verify" the client hostname.
     */
    else if (errno == 0 && (sa->sa_family == AF_INET
#ifdef INET6
			    || sa->sa_family == AF_INET6
#endif
	     )) {
#ifdef INET6
	char hbuf[NI_MAXHOST];
	char abuf[NI_MAXHOST];
	char rabuf[NI_MAXHOST];
	struct addrinfo hints, *res0 = NULL, *res;
	char *colonp;
#else
	char abuf[sizeof("255.255.255.255") + 1];
	char *hbuf;
#endif
	int error = -1;

#ifdef INET6
	error = getnameinfo(sa, len, abuf, sizeof(abuf), NULL, 0,
			    NI_NUMERICHOST | NI_WITHSCOPEID);
	if (error)
	    msg_fatal("%s: numeric getnameinfo lookup for peer: error %s",
		      myname, GAI_STRERROR(error));

	/*
	 * Convert an IPv4-mapped IPv6-address to 'true' IPv4 address
	 * early on. We have no need for the mapped form in logging,
	 * hostname verification and access checks.
	 */
	if (sa->sa_family == AF_INET6
	    && IN6_IS_ADDR_V4MAPPED(&((struct sockaddr_in6 *)sa)->sin6_addr)
	    && (colonp = strrchr(abuf, ':')) != NULL) {
	    struct addrinfo hints, *res0;
	    if (msg_verbose > 1)
		msg_info("%s: rewriting V4-mapped address \"%s\" to \"%s\"",
			 myname, abuf, colonp + 1);
	    state->addr = mystrdup(colonp + 1);
	    /*
	     * We create new socket information so getnameinfo() will be
	     * performed on the rewritten IPv4 address.
	     */
	    memset(&hints, 0, sizeof(hints));
	    hints.ai_family = AF_INET;
	    hints.ai_socktype = SOCK_STREAM;
	    hints.ai_flags = AI_NUMERICHOST;
	    error = getaddrinfo(state->addr, NULL, &hints, &res0);
	    if (error)
		msg_panic("%s: getaddrinfo(\"%s\", NULL, "
			  "{AF_INET,SOCK_STREAM,AI_NUMERICHOST}, "
			  "&res0): %s", myname, state->addr,
			  GAI_STRERROR(error));
	    len = res0->ai_addrlen;
	    memcpy((char *)sa, res0->ai_addr, len);
	} else {
	    state->addr = mystrdup(abuf);
	}

	/*
	 * RFC 2821 section 4.1.3: IPv6 address literals in SMTP
	 * mail headers are prepended with tag 'IPv6' and a colon.
	 */
	if (sa->sa_family == AF_INET6)
	    state->addr_tag = "IPv6:";

	error = getnameinfo(sa, len, hbuf, sizeof(hbuf), NULL, 0, NI_NAMEREQD);
#else
	in = &((struct sockaddr_in *)sa)->sin_addr;
	inet_ntop(AF_INET, in, abuf, sizeof(abuf));
	state->addr = mystrdup(abuf);
	hbuf = NULL;
	hp = gethostbyaddr((char *)in, sizeof(*in), AF_INET);
	if (hp) {
	    error = 0;
		hbuf = mystrdup(hp->h_name);
	} else
	    error = 1;
#endif
	if (error) {
	    state->name = mystrdup(CLIENT_NAME_UNKNOWN);
#ifdef INET6
	    if (error != EAI_NONAME)
		msg_warn("%s: getnameinfo(%s,,,,,,NI_NAMEREQD) error %s",
			 myname, abuf, GAI_STRERROR(error));
	    /*
	     * XXX: There are other error codes from GAI that should
	     * result in only a temporary error code from this daemon.
	     * This also applies to get{addr,name}info() results
	     * below.
	     */
	    state->peer_code = (error == EAI_AGAIN ?
				SMTPD_PEER_CODE_TEMP : SMTPD_PEER_CODE_PERM);
#else
	    state->peer_code = (h_errno == TRY_AGAIN ?
				SMTPD_PEER_CODE_TEMP : SMTPD_PEER_CODE_PERM);
#endif
	} else if (valid_hostaddr(hbuf, DONT_GRIPE)) {
	    msg_warn("numeric result %s in address->name lookup for %s",
		     hbuf, state->addr);
	    state->name = mystrdup(CLIENT_NAME_UNKNOWN);
	    state->peer_code = SMTPD_PEER_CODE_PERM;
	} else if (!valid_hostname(hbuf, DONT_GRIPE)) {
	    state->name = mystrdup(CLIENT_NAME_UNKNOWN);
	    state->peer_code = SMTPD_PEER_CODE_PERM;
	} else {
	    state->name = mystrdup(hbuf);
	    state->peer_code = SMTPD_PEER_CODE_OK;

	    /*
	     * Reject the hostname if it does not list the peer address.
	     */
#define REJECT_PEER_NAME(state, code) { \
	myfree(state->name); \
	state->name = mystrdup(CLIENT_NAME_UNKNOWN); \
	state->peer_code = code; \
    }

#ifdef INET6
	    memset(&hints, 0, sizeof(hints));
	    hints.ai_family = AF_UNSPEC;
	    hints.ai_socktype = SOCK_STREAM;
	    error = getaddrinfo(state->name, NULL, &hints, &res0);
	    if (error) {
		msg_warn("%s: %s: hostname %s verification failed: %s",
			 myname, state->addr, state->name,
			 GAI_STRERROR(error));
		REJECT_PEER_NAME(state, (error == EAI_AGAIN ?
				SMTPD_PEER_CODE_TEMP : SMTPD_PEER_CODE_PERM));
	    } else {
		for (res = res0; res; res = res->ai_next) {
		    if (res->ai_family != sa->sa_family)
			continue;
		    error = getnameinfo(res->ai_addr, res->ai_addrlen,
					rabuf, sizeof(rabuf), NULL, 0,
					NI_NUMERICHOST | NI_WITHSCOPEID);
		    if (error) {
			msg_warn("%s: %s: hostname %s verification failed: %s",
				 myname, state->addr, state->name,
				 GAI_STRERROR(error));
			REJECT_PEER_NAME(state, SMTPD_PEER_CODE_TEMP);
			break;
		    }
		    if (strcmp(state->addr, rabuf) == 0)
			break;	    /* keep peer name */
		}
		if (res == NULL) {
		    msg_warn("%s: %s: address not listed for hostname %s",
			     myname, state->addr, state->name);
		    REJECT_PEER_NAME(state, SMTPD_PEER_CODE_PERM);
		}
	    }
	    if (res0)
		freeaddrinfo(res0);
#else
	    hp = gethostbyname(state->name);
	    if (hp == 0) {
		msg_warn("%s: hostname %s verification failed: %s",
			 state->addr, state->name, HSTRERROR(h_errno));
		REJECT_PEER_NAME(state, (h_errno == TRY_AGAIN ?
			    SMTPD_PEER_CODE_TEMP : SMTPD_PEER_CODE_PERM));
	    } else if (hp->h_length != sizeof(*in)) {
		msg_warn("%s: hostname %s verification failed: bad address size %d",
			 state->addr, state->name, hp->h_length);
		REJECT_PEER_NAME(state, SMTPD_PEER_CODE_PERM);
	    } else {
		int i;
		for (i = 0; /* void */ ; i++) {
		    if (hp->h_addr_list[i] == 0) {
			msg_warn("%s: address not listed for hostname %s",
				 state->addr, state->name);
			REJECT_PEER_NAME(state, SMTPD_PEER_CODE_PERM);
			break;
		    }
		    if (memcmp(hp->h_addr_list[i], (char *)in, sizeof(*in)) == 0)
			break;			/* keep peer name */
		}
	    }
#endif
	}
    }

    /*
     * If it's not Internet, assume the client is local, and avoid using the
     * naming service because that can hang when the machine is disconnected.
     */
    else {
	state->name = mystrdup("localhost");
	state->addr = mystrdup("127.0.0.1");	/* XXX bogus. */
	state->peer_code = SMTPD_PEER_CODE_OK;
    }

    /*
     * Do the name[addr] formatting for pretty reports.
     */
    state->namaddr =
	concatenate(state->name, "[", state->addr, "]", (char *) 0);
}

/* smtpd_peer_reset - destroy peer information */

void    smtpd_peer_reset(SMTPD_STATE *state)
{
    myfree(state->name);
    myfree(state->addr);
    myfree(state->namaddr);
}
