/*++
/* NAME
/*	inet_addr_local 3
/* SUMMARY
/*	determine if IP address is local
/* SYNOPSIS
/*	#include <inet_addr_local.h>
/*
/*	int	inet_addr_local(addr_list, mask_list, addr_family)
/*	INET_ADDR_LIST *addr_list;
/*	INET_ADDR_LIST *mask_list;
/*	int addr_family;
/* DESCRIPTION
/*	inet_addr_local() determines all active IP interface addresses
/*	of the local system. Any address found is appended to the
/*	specified address list. The result value is the number of
/*	active interfaces found.
/*
/*	The mask_list is either a null pointer, or it is a list that
/*	receives the netmasks of the interface addresses that were found.
/*	
/*	The addr_family is ether AF_UNSPEC, AF_INET or AF_INET6
/*
/* DIAGNOSTICS
/*	Fatal errors: out of memory.
/* SEE ALSO
/*	inet_addr_list(3) address list management
/* LICENSE
/* .ad
/* .fi
/*	The Secure Mailer license must be distributed with this software.
/* AUTHOR(S)
/*	Wietse Venema
/*	IBM T.J. Watson Research
/*	P.O. Box 704
/*	Yorktown Heights, NY 10598, USA
/*
/*	Dean C. Strik
/*	Department ICT
/*	Eindhoven University of Technology
/*	P.O. Box 513
/*	5600 MB  Eindhoven, Netherlands
/*	E-mail: <dean@ipnet6.org>
/*--*/

/* System library. */

#include <sys_defs.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <arpa/inet.h>
#include <unistd.h>
#ifdef USE_SYS_SOCKIO_H
#include <sys/sockio.h>
#endif
#include <errno.h>
#include <string.h>
#ifdef INET6
#include <netdb.h>
#include <stdio.h>
#endif
#ifdef HAVE_GETIFADDRS
#include <ifaddrs.h>
#endif

/* Utility library. */

#include <msg.h>
#include <mymalloc.h>
#include <vstring.h>
#include <inet_addr_list.h>
#include <inet_addr_local.h>

 /*
  * IF IPV6 SUPPORT IS ENABLED:
  *
  * In the non-getifaddrs() version, we determine the interface addresses
  * using the SIOCG(L)IFCONF. However, it is operating system dependent
  * whether this also results in IPv6 addresses configuration. Another
  * issue is that there is no good method to determine the netmask /
  * prefixlen for IPv6 addresses.
  * We will therefore use OS dependent methods. An overview:
  *  - Use SIOCGLIFCONF when available -> this supports both IPv4/IPv6
  *    addresses. Also, with SIOCGLIFNETMASK we can obtain the netmask /
  *    prefixlen for either address family.
  *  - On Linux, read IPv6 addresses / prefixlengths from a file in the
  *    /proc filesystem. Linux does not return IPv6 addresses in
  *    SIOCGIFCONF.
  *  - On other systems without getifaddrs(), we expect SIOCGIFCONF
  *    to return IPv6 addresses. Since SIOCGIFNETMASK does not work for
  *    IPv6 addresses, we will always set the prefixlen to 128 (host-only)
  *    However, it is suggested you set the mynetworks variable(s)
  *    manually then.
  *    XXX: We duplicate some code. In this case, I think this is better
  *    than really drowning in the #ifdefs...
  * -- Dean Strik (dcs)
  */

 /*
  * Support for variable-length addresses.
  */
#ifdef HAS_SIOCGLIF
#else /* HAS_SIOCGLIF */
#endif /* HAS_SIOCGLIF */

/* decode_scope - separate scope ID from IPv6 address */

#ifdef INET6
static struct sockaddr *decode_scope(struct sockaddr *sa,
				     struct sockaddr_in6 *sin6)
{
#ifdef INET6_KAME
    memcpy(sin6, sa, sa->sa_len);	/* size sin6 >> size sa */
    /* decode scoped address notation */
    if ((IN6_IS_ADDR_LINKLOCAL(&sin6->sin6_addr) ||
	    IN6_IS_ADDR_SITELOCAL(&sin6->sin6_addr)) &&
	    sin6->sin6_scope_id == 0) {
	sin6->sin6_scope_id = ntohs(*(u_int16_t *)&sin6->sin6_addr.s6_addr[2]);
	sin6->sin6_addr.s6_addr[2] = sin6->sin6_addr.s6_addr[3] = 0;
    }
    return (struct sockaddr *)sin6;
#else
    return (sa);
#endif
}
#endif

/* ial_socket - make socket for ioctl() operations */

static int ial_socket(int af)
{
    char  *myname = "inet_addr_local[socket]";
    int    sock;

    /*
     * The host may not be actually configured with IPv6. When
     * IPv6 support is not actually in the kernel, don't consider
     * failure to create an IPv6 socket as fatal. This could be
     * tuned better though. For other families, the error is fatal.
     */
    if ((sock = socket(af, SOCK_DGRAM, 0)) < 0) {
#ifdef INET6
	if (af == AF_INET6) {
	    if (msg_verbose)
		msg_warn("%s: socket: %m", myname);
	    return (-1);
	}
#endif
	msg_fatal("%s: socket: %m", myname);
    }
    return (sock);
}


#ifdef HAVE_GETIFADDRS

/*
 * The getifaddrs(3) function, introduced by BSD/OS, provides a
 * platform-independent way of requesting interface addresses,
 * including IPv6 addresses. The implementation however is not
 * present in all major operating systems.
 */

/* ial_getifaddrs - determine IP addresses using getifaddrs(3) */

static int ial_getifaddrs(INET_ADDR_LIST *addr_list,
			  INET_ADDR_LIST *mask_list,
			  int af)
{
    char *myname = "inet_addr_local[getifaddrs]";
    struct ifaddrs *ifap, *ifa;
    struct sockaddr *sa, *sam;
#ifdef INET6
    struct sockaddr_in6 addr6;
#else
    void *addr,*addrm;
#endif

    if (getifaddrs(&ifap) < 0)
	msg_fatal("%s: getifaddrs: %m", myname);

    /*
     * Get the address of each IP network interface. According to BIND we
     * must include interfaces that are down because the machine may still
     * receive packets for that address (yes, via some other interface).
     * Having no way to verify this claim on every machine, I will give them
     * the benefit of the doubt.
     */

    for (ifa = ifap; ifa; ifa = ifa->ifa_next) {
	if (!(ifa->ifa_flags & IFF_RUNNING) || ifa->ifa_addr == NULL) 
	    continue;
	sa = ifa->ifa_addr;
	sam = ifa->ifa_netmask;
	if (af != AF_UNSPEC && sa->sa_family != af)
	    continue;
	switch (sa->sa_family) {
	case AF_INET:
#ifndef INET6
	    addr = (void *)&((struct sockaddr_in *)sa)->sin_addr;
	    addrm = (void *)&((struct sockaddr_in *)ifa->ifa_netmask)->sin_addr;
#endif
	    break;
#ifdef INET6
	case AF_INET6:
	    sa = decode_scope(sa, &addr6);
	    break;
#endif
	default:
	    continue;
	}

#ifdef INET6
	inet_addr_list_append(addr_list, sa);
	if (mask_list != NULL) {
	    /*
	     * Unfortunately, sa_len/sa_family may be broken in
	     * the netmask sockaddr structure. We must fix this
	     * manually to have correct addresses.   --dcs
	     */
#ifdef HAS_SA_LEN
	    sam->sa_len = sa->sa_family == AF_INET6 ?
			  sizeof(struct sockaddr_in6) :
			  sizeof(struct sockaddr_in);
#endif
	    sam->sa_family = sa->sa_family;
	    inet_addr_list_append(mask_list, sam);
	}
#else
	inet_addr_list_append(addr_list, (struct in_addr *)addr);
	if (mask_list != NULL)
	    inet_addr_list_append(mask_list, (struct in_addr *)addrm);
#endif
    }

    freeifaddrs(ifap);
    return (0);
}
#endif /* HAVE_GETIFADDRS */


#ifdef HAS_SIOCGLIF

/*
 * The SIOCLIF* ioctls are the successors of SIOCGIF* on the Solaris
 * and HP/UX operating systems. The data is stored in sockaddr_storage
 * structure. Both IPv4 and IPv6 addresses are returned though these
 * calls.
 */
#define NEXT_INTERFACE(lifr) (lifr + 1)
#define LIFREQ_SIZE(lifr) sizeof(lifr[0])
#define ial_generic ial_siocglif

/* ial_siocglif - determine IP addresses using ioctl(SIOCGLIF*) */

static int ial_siocglif(INET_ADDR_LIST *addr_list,
			INET_ADDR_LIST *mask_list,
			int af)
{
    char *myname = "inet_addr_local[siocglif]";
    struct lifconf lifc;
    struct lifreq *lifr;
    struct lifreq *lifr_mask;
    struct lifreq *the_end;
    struct sockaddr *sa;
    struct sockaddr_in6 addr6;
    int   sock;
    VSTRING *buf;

    if (af != AF_INET && af != AF_INET6)
	msg_fatal("%s: address family was %d, must be AF_INET (%d) or "
		  "AF_INET6 (%d)", myname, af, AF_INET, AF_INET6);
    sock = ial_socket(af);
    if (sock < 0)
	return (0);
    buf = vstring_alloc(1024);
    for (;;) {
	memset(&lifc, 0, sizeof(lifc));
	lifc.lifc_family = AF_UNSPEC;
	lifc.lifc_len = vstring_avail(buf);
	lifc.lifc_buf = vstring_str(buf);
	if (ioctl(sock, SIOCGLIFCONF, (char *) &lifc) < 0) {
	    if (errno != EINVAL)
		msg_fatal("%s: ioctl SIOCGLIFCONF: %m", myname);
	} else if (lifc.lifc_len < vstring_avail(buf) / 2)
	    break;
	VSTRING_SPACE(buf, vstring_avail(buf) * 2);
    }

    the_end = (struct lifreq *) (lifc.lifc_buf + lifc.lifc_len);
    for (lifr = lifc.lifc_req; lifr < the_end;) {
	if (((struct sockaddr *)&lifr->lifr_addr)->sa_family != af) {
	    lifr = NEXT_INTERFACE(lifr);
	    continue;
	}
	if (af == AF_INET) {
	    if (((struct sockaddr_in *)&lifr->lifr_addr)->sin_addr.s_addr
		    == INADDR_ANY) {
		lifr = NEXT_INTERFACE(lifr);
		continue;
	    }
	    sa = (struct sockaddr *)&lifr->lifr_addr;
	} else if (af == AF_INET6) {
	    sa = decode_scope((struct sockaddr *)&lifr->lifr_addr, &addr6);
	    if (IN6_IS_ADDR_UNSPECIFIED(&addr6.sin6_addr)) {
		lifr = NEXT_INTERFACE(lifr);
		continue;
	    }
	}
	inet_addr_list_append(addr_list, sa);
	if (mask_list) {
	    lifr_mask = (struct lifreq *) mymalloc(sizeof(struct lifreq));
	    memcpy((char *)lifr_mask, (char *)lifr, sizeof(struct lifreq));
	    if (ioctl(sock, SIOCGLIFNETMASK, lifr_mask) < 0)
		msg_fatal("%s: ioctl(SIOCGLIFNETMASK): %m", myname);
	    /* XXX: Check whether sa_len/family are honoured --dcs */
	    inet_addr_list_append(mask_list,
				 (struct sockaddr *)&lifr_mask->lifr_addr);
	    myfree((char *)lifr_mask);
	}
	lifr = NEXT_INTERFACE(lifr);
    }
    vstring_free(buf);
    (void) close(sock);
    return (0);
}

#else /* HAVE_SIOCGLIF */

/*
 * The classic SIOCGIF* ioctls. Modern BSD operating systems will
 * also return IPv6 addresses through these structure. Note however
 * that recent versions of these operating systems have getifaddrs.
 */
#define ial_generic ial_siocgif
#ifdef _SIZEOF_ADDR_IFREQ
#define NEXT_INTERFACE(ifr) ((struct ifreq *) \
	((char *) ifr + _SIZEOF_ADDR_IFREQ(*ifr)))
#define IFREQ_SIZE(ifr)	_SIZEOF_ADDR_IFREQ(*ifr)
#else /* _SIZEOF_ADDR_IFREQ */
#ifdef HAS_SA_LEN
#define NEXT_INTERFACE(ifr) ((struct ifreq *) \
	((char *) ifr + sizeof(ifr->ifr_name) + ifr->ifr_addr.sa_len))
#define IFREQ_SIZE(ifr)	(sizeof(ifr->ifr_name) + ifr->ifr_addr.sa_len)
#else /* HAS_SA_LEN */
#define NEXT_INTERFACE(ifr) (ifr + 1)
#define IFREQ_SIZE(ifr)	sizeof(ifr[0])
#endif /* HAS_SA_LEN */
#endif /* _SIZEOF_ADDR_IFREQ */

/* ial_siocgif - determine IP addresses using ioctl(SIOCGIF*) */

static int ial_siocgif(INET_ADDR_LIST *addr_list,
			INET_ADDR_LIST *mask_list,
			int af)
{
    char *myname = "inet_addr_local[siocgif]";
    struct in_addr addr;
    struct ifconf ifc;
    struct ifreq *ifr;
    struct ifreq *ifr_mask;
    struct ifreq *the_end;
#ifdef INET6
    struct sockaddr *sa;
    struct sockaddr_in6 addr6;
#endif
    int   sock;
    VSTRING *buf;

    /*
     * Get the network interface list. XXX The socket API appears to have no
     * function that returns the number of network interfaces, so we have to
     * guess how much space is needed to store the result.
     * 
     * On BSD-derived systems, ioctl SIOCGIFCONF returns as much information as
     * possible, leaving it up to the application to repeat the request with
     * a larger buffer if the result caused a tight fit.
     * 
     * Other systems, such as Solaris 2.5, generate an EINVAL error when the
     * buffer is too small for the entire result. Workaround: ignore EINVAL
     * errors and repeat the request with a larger buffer. The downside is
     * that the program can run out of memory due to a non-memory problem,
     * making it more difficult than necessary to diagnose the real problem.
     */

    sock = ial_socket(af);
    if (sock < 0)
	return (0);
    buf = vstring_alloc(1024);
    for (;;) {
	ifc.ifc_len = vstring_avail(buf);
	ifc.ifc_buf = vstring_str(buf);
	if (ioctl(sock, SIOCGIFCONF, (char *) &ifc) < 0) {
	    if (errno != EINVAL)
		msg_fatal("%s: ioctl SIOCGIFCONF: %m", myname);
	} else if (ifc.ifc_len < vstring_avail(buf) / 2)
	    break;
	VSTRING_SPACE(buf, vstring_avail(buf) * 2);
    }

    the_end = (struct ifreq *) (ifc.ifc_buf + ifc.ifc_len);
    for (ifr = ifc.ifc_req; ifr < the_end;) {
        if (ifr->ifr_addr.sa_family != af) {
	    ifr = NEXT_INTERFACE(ifr);
	    continue;
        }
	if (af == AF_INET) {
	    addr = ((struct sockaddr_in *) &ifr->ifr_addr)->sin_addr;
	    if (addr.s_addr != INADDR_ANY) {
#ifdef INET6
		inet_addr_list_append(addr_list, &ifr->ifr_addr);
#else
		inet_addr_list_append(addr_list, &addr);
#endif
		if (mask_list) {
		    ifr_mask = (struct ifreq *) mymalloc(IFREQ_SIZE(ifr));
		    memcpy((char *) ifr_mask, (char *) ifr, IFREQ_SIZE(ifr));
		    if (ioctl(sock, SIOCGIFNETMASK, ifr_mask) < 0)
			msg_fatal("%s: ioctl SIOCGIFNETMASK: %m", myname);
#ifdef INET6
		    /*
		     * Note that this SIOCGIFNETMASK has truly screwed up
		     * the contents of sa_len/sa_family. We must fix this
		     * manually to have correct addresses.   --dcs
		     */
#ifdef HAS_SA_LEN
		    ifr_mask->ifr_addr.sa_len = sizeof(struct sockaddr_in);
#endif
		    ifr_mask->ifr_addr.sa_family = af;
		    inet_addr_list_append(mask_list, &ifr_mask->ifr_addr);
#else
		    addr = ((struct sockaddr_in *) &ifr_mask->ifr_addr)->sin_addr;
		    inet_addr_list_append(mask_list, &addr);
#endif
		    myfree((char *) ifr_mask);
		}
	    }
	}
#ifdef INET6
	else if (af == AF_INET6) {
	    sa = decode_scope(&ifr->ifr_addr, &addr6);
	    if (!(IN6_IS_ADDR_UNSPECIFIED(&addr6.sin6_addr))) {
	        inet_addr_list_append(addr_list, sa);
		if (mask_list) {
		    /* We can't know, and assume /128 for everything */
		    struct sockaddr_in6 mask6;
		    memcpy((char *)&mask6, (char *)&addr6,
			   sizeof(struct sockaddr_in6));
		    memset((char *)&mask6.sin6_addr, ~0,
			   sizeof(mask6.sin6_addr));
		    inet_addr_list_append(mask_list,
					  (struct sockaddr *)&mask6);
		}
	    }
	}
#endif /* INET6 */
	ifr = NEXT_INTERFACE(ifr);
    }
    vstring_free(buf);
    (void) close(sock);
    return (0);
}
#endif /* HAVE_SIOCGLIF */


#ifdef HAS_PROCNET_IFINET6

/*
 * Linux does not provide proper calls to retrieve IPv6 interface
 * addresses. Instead, the addresses can be read from a file in the
 * /proc tree. The most important issue with this approach however
 * is that the /proc tree may not always be available, for example
 * in a chrooted environment or in "hardened" (sic) installations.
 */

/* ial_procnet_ifinet6 - determine IPv6 addresses using /proc/net/if_inet6 */

static int ial_procnet_ifinet6(INET_ADDR_LIST *addr_list,
			       INET_ADDR_LIST *mask_list)
{
    char *myname = "inet_addr_local[procnet_ifinet6]";
    FILE *f;
    char addr6p[8][5], addr6res[40], devname[20];
    int plen, scope, dad_status, if_idx, gaierror;
    struct addrinfo hints, *res, *res0;

    if ((f = fopen(_PATH_PROCNET_IFINET6, "r")) != NULL) {
	while (fscanf(f, "%4s%4s%4s%4s%4s%4s%4s%4s %02x %02x %02x %02x %20s\n",
		addr6p[0], addr6p[1], addr6p[2], addr6p[3], addr6p[4],
		addr6p[5], addr6p[6], addr6p[7],
		&if_idx, &plen, &scope, &dad_status, devname) != EOF) {
	    sprintf(addr6res, "%s:%s:%s:%s:%s:%s:%s:%s",
		addr6p[0], addr6p[1], addr6p[2], addr6p[3],
		addr6p[4], addr6p[5], addr6p[6], addr6p[7]);
	    addr6res[sizeof(addr6res) - 1] = 0;
	    memset(&hints, 0, sizeof(hints));
	    hints.ai_flags = AI_NUMERICHOST;
	    hints.ai_family = AF_INET6;
	    hints.ai_socktype = SOCK_DGRAM;
	    gaierror = getaddrinfo(addr6res, NULL, &hints, &res0);
	    if (!gaierror) {
		for (res = res0; res; res = res->ai_next) {
		    struct sockaddr_in6 mask;
		    int i, rest;
		    inet_addr_list_append(addr_list, res->ai_addr);
		    memcpy((char *)&mask, res->ai_addr, res->ai_addrlen);
		    /* s6_addr32 is available on linux */
		    mask.sin6_addr.s6_addr32[0] =
		    mask.sin6_addr.s6_addr32[1] =
		    mask.sin6_addr.s6_addr32[2] =
		    mask.sin6_addr.s6_addr32[3] = ~0;
		    for (i = 3, rest = 128 - plen; i > -1; i--)
			if (rest > 31) {
			    mask.sin6_addr.s6_addr32[i] = htonl(0);
			    rest -= 32;
			} else {
			    mask.sin6_addr.s6_addr32[i] =
				htonl(~((1 << rest) - 1));
			    break;
			}
		    inet_addr_list_append(mask_list, (struct sockaddr *)&mask);
		}
		freeaddrinfo(res0);
	    }
	}
    } else if (msg_verbose) {
	msg_warn("%s: Couldn't open " _PATH_PROCNET_IFINET6
		 " for reading: %m", myname);
    }
    return (0);
}
#endif /* HAS_PROCNET_IFINET6 */


/* inet_addr_local - find all IP addresses for this host */

int     inet_addr_local(INET_ADDR_LIST *addr_list, INET_ADDR_LIST *mask_list,
			int addr_family)
{
    char   *myname = "inet_addr_local";
    int     initial_count = addr_list->used;
    int     count;

    /*
     * IP Version 4
     */
    if (addr_family == AF_INET || addr_family == AF_UNSPEC) {
	count = addr_list->used;
#if defined(HAVE_GETIFADDRS)
	ial_getifaddrs(addr_list, mask_list, AF_INET);
#else
	ial_generic(addr_list, mask_list, AF_INET);
#endif
	if (msg_verbose)
	    msg_info("%s: configured %d IPv4 addresses",
		     myname, addr_list->used - count);
    }

    /*
     * IP Version 6
     */
    if (addr_family == AF_INET6 || addr_family == AF_UNSPEC) {
	count = addr_list->used;
#ifdef INET6
#if defined(HAS_PROCNET_IFINET6)
	ial_procnet_ifinet6(addr_list, mask_list);
#elif defined(HAVE_GETIFADDRS)
	ial_getifaddrs(addr_list, mask_list, AF_INET6);
#else
	ial_generic(addr_list, mask_list, AF_INET6);
#endif
	if (msg_verbose)
	    msg_info("%s: configured %d IPv6 addresses", myname,
		     addr_list->used - count);
#endif
    }

    return (addr_list->used - initial_count);
}


#ifdef TEST
/* XXX: Requires INET6 for now */

#include <string.h>
#include <vstream.h>
#include <msg_vstream.h>

int     main(int unused_argc, char **argv)
{
    INET_ADDR_LIST addr_list;
    INET_ADDR_LIST mask_list;
    int     i;
    char abuf[NI_MAXHOST], mbuf[NI_MAXHOST];
    struct sockaddr *sa;

    msg_vstream_init(argv[0], VSTREAM_ERR);

    inet_addr_list_init(&addr_list);
    inet_addr_list_init(&mask_list);
    inet_addr_local(&addr_list, &mask_list, AF_UNSPEC);

    if (addr_list.used == 0)
	msg_fatal("cannot find any active network interfaces");

    if (addr_list.used == 1)
	msg_warn("found only one active network interface");

    for (i = 0; i < addr_list.used; i++) {
	sa = (struct sockaddr *)&addr_list.addrs[i];
	if (getnameinfo(sa, SA_LEN(sa), abuf, sizeof(abuf), NULL, 0,
		NI_NUMERICHOST)) {
	    strncpy(abuf, "???", sizeof(abuf));
	}
	sa = (struct sockaddr *)&mask_list.addrs[i];
	if (getnameinfo(sa, SA_LEN(sa), mbuf, sizeof(mbuf), NULL, 0,
		NI_NUMERICHOST)) {
	    strncpy(mbuf, "???", sizeof(mbuf));
	}
	vstream_printf("%s/%s\n", abuf, mbuf);
    }
    vstream_fflush(VSTREAM_OUT);
    inet_addr_list_free(&addr_list);
    inet_addr_list_free(&mask_list);
}

#endif
